/*
 * Simplified OLTP Simulation.
 *
 * This program tries to reproduce the OLTP benchmark as defined by the
 * filebench utility, using threads instead fo processes.
 */
#define _XOPEN_SOURCE 500
#define _GNU_SOURCE
#include <ctype.h>
#include <fcntl.h>
#include <pthread.h>
#include <libaio.h>
#include <unistd.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#define MAX_FILENAME		256
#define BLOCK_SIZE		512
#define READ_SIZE		4096

#define DEFAULT_NREADERS	100
#define DEFAULT_NWRITERS	10
#define DEFAULT_BASEDIR		"/mnt/hdb1/"
#define DEFAULT_OUTDIR		"./"
#define DEFAULT_SIZE		(10 * 1024 * 1024)
#define DEFAULT_DURATION	600
#define DEFAULT_DEPTH		1

static int async = 1;
static int nreaders = DEFAULT_NREADERS;
static int nwriters = DEFAULT_NWRITERS;
static char *basedir = DEFAULT_BASEDIR;
static char *outdir = DEFAULT_OUTDIR;
static int sectors = DEFAULT_SIZE;

#define CREATE			1
#define RUN			2

#define READ			1
#define WRITE			2

static int mode;
static int dbfd;
static int stop_all;
static intptr_t page_mask;

#define SEQ_ENTRIES	2048

struct log_seq {
	unsigned delay[SEQ_ENTRIES];
	int free;
	struct log_seq *next;
};

struct worker {
	pthread_t id;
	pthread_mutex_t startup;
	char buffer[2*READ_SIZE*DEFAULT_DEPTH];
	char *bufptr[DEFAULT_DEPTH];
	void (*loop_init)(struct worker *worker);
	void (*loop_body)(struct worker *worker);
	struct log_seq seq;

	int ddir;

	io_context_t aio_ctx;
	int submitted;
	int busy[DEFAULT_DEPTH];
	struct iocb iocbs[DEFAULT_DEPTH];
	struct timeval subtime[DEFAULT_DEPTH];
	struct io_event event;
};

//static struct worker *logwriter;
static struct worker *readers;
static struct worker *writers;

static struct worker **all_workers[] = { &readers, &writers, NULL };
static int *num_workers[] = { &nreaders, &nwriters };

#define for_all_workers(r, i, j)					\
	for (j = 0; all_workers[j]; j++)				\
		for (r = *all_workers[j], i = 0; i < *num_workers[j]; r++, i++)

static void bad_usage(void)
{
	printf("usage: oltp [-r readers] [-w writers] [-d dir] run\n"
	       "            [-d dir] [-s size] create\n");
	exit(-1);
}

#define error(msg...) do {						\
	printf("%s:%d: ", __FUNCTION__, __LINE__);			\
	printf(msg);							\
	exit(-1);							\
} while (0)

static void csnprintf(char *buf, int size, const char *base, const char *name)
{
	int r;

	r = snprintf(buf, size, "%s/%s", base, name);
	if (r == size)
		error("snprintf failed\n");
}

static void *cmalloc(int size)
{
	void *ptr;

	ptr = calloc(1, size);
	if (!ptr)
		error("calloc failed\n");
	return ptr;
}

static void parse_options(int argc, char *argv[])
{
	int c;

	while ((c = getopt(argc, argv, "d:r:w:s:")) != -1) {
		switch (c) {
		case 'r':
			nreaders = strtol(optarg, NULL, 10);
			break;
		case 'w':
			nwriters = strtol(optarg, NULL, 10);
			break;
		case 'd':
			basedir = optarg;
			break;
		case 's':
			sectors = strtol(optarg, NULL, 10);
			break;
		default:
			bad_usage();
		}
	}

	if (optind != argc - 1)
		bad_usage();

	if (!strcmp(argv[optind], "run"))
		mode = RUN;
	else if (!strcmp(argv[optind], "create"))
		mode = CREATE;
	else
		bad_usage();
}

static void seq_dump(FILE *file, struct log_seq *seq)
{
	int i;

	while (seq) {
		for (i = 0; i < seq->free; i++)
			fprintf(file, "%u\n", seq->delay[i]);
		seq = seq->next;
	}
}

static void seq_append(struct log_seq *seq, unsigned delay)
{
	if (seq->free == SEQ_ENTRIES) {
		seq->next = cmalloc(sizeof(struct log_seq));
		seq = seq->next;
	}

	seq->delay[seq->free++] = delay;
}

static off_t brand(off_t a, off_t b)
{
	off_t x;

	x = (off_t)((b - a) * (float)rand() / RAND_MAX);
	return x + a;
}

static unsigned calc_delta(struct timeval *tv0, struct timeval *tv1)
{
	return (tv0->tv_sec - tv1->tv_sec) * 1000000 +
		tv0->tv_usec - tv1->tv_usec;
}

static void async_loop_init(struct worker *worker)
{
	int r;

	r = io_queue_init(DEFAULT_DEPTH, &worker->aio_ctx);
	if (r)
		error("io_queue_init failed\n");
}

static int alloc_iocb(struct worker *worker)
{
	int i;

	for (i = 0; i < DEFAULT_DEPTH; i++) {
		if (!worker->busy[i]) {
			worker->busy[i] = 1;
			return i;
		}
	}

	return -1;
}

static void async_loop_body(struct worker *worker)
{
	struct timespec to = { .tv_sec = 1, .tv_nsec = 0 };
	struct timeval tv0;
	struct iocb *iocb, *iocbv[1];
	off_t offset;
	int idx, r;

	while (worker->submitted < DEFAULT_DEPTH) {
		offset = (off_t)brand(0, sectors - 2 * READ_SIZE /
				      BLOCK_SIZE) * BLOCK_SIZE;
		offset = (offset + page_mask) & ~page_mask;
		idx = alloc_iocb(worker);

		if (worker->ddir == READ) {
			io_prep_pread(&worker->iocbs[idx], dbfd,
				      worker->bufptr[idx], READ_SIZE, offset);
		} else {
			io_prep_pwrite(&worker->iocbs[idx], dbfd,
				      worker->bufptr[idx], READ_SIZE, offset);
		}

		gettimeofday(&worker->subtime[idx], NULL);
		iocbv[0] = &worker->iocbs[idx];
		r = io_submit(worker->aio_ctx, 1, iocbv);
		if (r <= 0)
			error("io_submit failed\n");

		worker->submitted++;
	}

	r = io_getevents(worker->aio_ctx, 1, 1, &worker->event, &to);
	if (r == 1) {
		iocb = worker->event.obj;
		idx = iocb - worker->iocbs;
		gettimeofday(&tv0, NULL);
		seq_append(&worker->seq, calc_delta(&tv0,
			   &worker->subtime[idx]));
		if (worker->event.res > READ_SIZE)
			error("async error\n");
		worker->submitted--;
		worker->busy[idx] = 0;
	}

	if (worker->ddir == WRITE) {
		/* thinktime */
		usleep(1000000);
	}
}

static void sync_loop_body(struct worker *worker)
{
	struct timeval tv0, tv1;
	off_t offset;
	int r;

	offset = (off_t)brand(0, sectors - 2 * READ_SIZE /
			      BLOCK_SIZE) * BLOCK_SIZE;
	offset = (offset + page_mask) & ~page_mask;
	gettimeofday(&tv0, NULL);
	if (worker->ddir == READ)
		r = pread(dbfd, worker->bufptr[0], READ_SIZE, offset);
	else
		r = pwrite(dbfd, worker->bufptr[0], READ_SIZE, offset);
	gettimeofday(&tv1, NULL);
	if (r < 0)
		error("read/write failed\n");
	seq_append(&worker->seq, calc_delta(&tv1, &tv0));
	if (worker->ddir == WRITE) {
		/* thinktime */
		usleep(1000000);
	}
}

static void *worker_body(void *arg)
{
	struct worker *worker = arg;

	if (worker->loop_init)
		worker->loop_init(worker);

	pthread_mutex_lock(&worker->startup);

	while (!stop_all)
		worker->loop_body(worker);

	return NULL;
}

static int create_files(void)
{
	char name[MAX_FILENAME];
	char block[BLOCK_SIZE];
	int i, fd;

	csnprintf(name, MAX_FILENAME, basedir, "dbfile");
	memset(block, 0, BLOCK_SIZE);

	fd = open(name, O_WRONLY | O_TRUNC | O_CREAT, S_IRUSR | S_IWUSR);
	for (i = 0; i < sectors; i++)
		write(fd, block, BLOCK_SIZE);
	close(fd);

	return 0;
}

static void stop_workers(void)
{
	struct worker *worker;
	int i, j;

	stop_all = 1;
	for_all_workers(worker, i, j)
		pthread_join(worker->id, NULL);
}

static struct worker *alloc_workers(int n, int async, int ddir)
{
	struct worker *workers, *worker;
	int i, j, r;

	workers = cmalloc(n * sizeof(struct worker));
	for (worker = workers, i = 0; i < n; worker++, i++) {
		pthread_mutex_init(&worker->startup, NULL);
		pthread_mutex_lock(&worker->startup);

		for (j = 0; j < DEFAULT_DEPTH; j++) {
			worker->bufptr[j] = (char *)
				(((uintptr_t)worker->buffer +
				2*READ_SIZE * j + page_mask) & ~page_mask);
		}
		if (async) {
			worker->loop_init = async_loop_init;
			worker->loop_body = async_loop_body;
		} else
			worker->loop_body = sync_loop_body;
		worker->ddir = ddir;

		r = pthread_create(&worker->id, NULL, worker_body, worker);
		if (r)
			error("pthread_create failed\n");
	}

	return workers;
}

static void dump_latencies(const char *fn, struct worker *array, int n)
{
	struct worker *worker;
	FILE *outfile;
	char name[MAX_FILENAME];
	int i;

	csnprintf(name, MAX_FILENAME, outdir, fn);
	outfile = fopen(name, "w");
	if (!outfile)
		error("fopen failed\n");
	for (worker = array, i = 0; i < n; worker++, i++)
		seq_dump(outfile, &worker->seq);
	fclose(outfile);
}

static int run(void)
{
	struct worker *worker;
	char name[MAX_FILENAME];
	int i, r;

	page_mask = sysconf(_SC_PAGESIZE);
	if (page_mask-- < 0)
		error("sysconf failed\n");

	csnprintf(name, MAX_FILENAME, basedir, "dbfile");
	dbfd = open(name, O_RDWR | O_DIRECT | O_SYNC);
	if (dbfd < 0)
		error("open failed\n");

	readers = alloc_workers(nreaders, async, READ);
	writers = alloc_workers(nwriters, async, WRITE);

	for_all_workers(worker, i, r)
		pthread_mutex_unlock(&worker->startup);

	sleep(DEFAULT_DURATION);

	stop_workers();
	dump_latencies("readers_clat.log", readers, nreaders);
	dump_latencies("writers_clat.log", writers, nwriters);

	return 0;
}

int main(int argc, char *argv[])
{
	parse_options(argc, argv);

	if (mode == CREATE)
		return create_files();
	else if (mode == RUN)
		return run();

	return -1;
}
