#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <string.h>

const int BLOCK_SIZE  = 4096 ;	                /* 4kB reads */
const int HTML_FILE_SIZE = 16*1024 ;            /* 16KiB file */
const int LARGE_MIN_FILE_SIZE = 1*1024*1024 ;
const int LARGE_MAX_FILE_SIZE = 30*1024*1024 ;
const int MAX_APPEND_SIZE = 16*1024 ;

const int HTML_FILES_PER_DIR = 50000 ;
const int LARGE_FILES_PER_DIR = 30 ;

const int PROCESS_NR = 100 ;
const int FILES_PER_PROCESS = 10 ;

const int MAX_PER_PROCESS_ITER = 50 ;

const int MAXLEN = 128 ;

const int PROB_SCALE = 10 ;
const int LARGE_FILE_PROB = 1 ;

const char *html_lat_log_basefilename = "webserver_emu_html_lat.log" ;
FILE* html_lat_log_fp ;

const char *large_bw_log_basefilename = "webserver_emu_large_bw.log" ;
FILE* large_bw_log_fp ;

//#define SIMULATED
#ifdef SIMULATED

const int DIR_NR = 2 ;
const int partition_numbers[DIR_NR] = { 1, 12} ;
const char dir_basename[] = "./hdb" ;
const char fakelog_filename[] = "./hdbXXX/webserver_emu.fakelog" ;

#else

const int HTML_DIR_NR = 16 ;
const int html_partition_numbers[HTML_DIR_NR] = 
    { 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34} ;
    // { 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32, 34} ;
const int LARGE_DIR_NR = 14 ;
const int large_partition_numbers[LARGE_DIR_NR] = 
    { 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18} ;
    //{ 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31, 33} ;
const char dir_basename[] = "/mnt/hdb" ;
const char fakelog_filename[] = "/mnt/hdb22/webserver_emu.fakelog" ;

#endif

const char html_file_basename[] = "webserver_emu.fakehtml." ;
const char large_file_basename[] = "webserver_emu.largefile" ;

char html_dirnames[HTML_DIR_NR][MAXLEN] ;
char large_dirnames[LARGE_DIR_NR][MAXLEN] ;

void check_rewrite_file(const char filename[MAXLEN], int target_size) 
{
    unsigned char *buffer;
    struct stat file_stat ;
    int failure ;

    failure = stat(filename, &file_stat) < 0 ;
    if (failure && errno != ENOENT) {
	printf("Unexpected error while retrieving information for "
	       "file %s\n", filename);
	exit(1) ;
    }

    // if file not found or size too small
    if ((failure && errno == ENOENT) || 
	file_stat.st_size < target_size) {
	buffer = (unsigned char *)malloc(sizeof(unsigned char) * target_size) ;
	if (buffer == NULL) {
	    printf("could not create write buffer");
	    exit(1);
	}

	int fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR) ;
	if (fd == -1) {
	    printf("could not open/create %s: %m\n", filename);
	    exit(1);
	}
	if (write(fd, buffer, target_size) <= 0) {
	    printf("could not write to file %s: %m\n", filename);
	    exit(1);
	}
 	free(buffer) ;
	fprintf(stderr, "created/rewrote %s ", filename) ;
	close(fd) ;
    } // end of file creation/rewrite case
} 

void check_create_files(const int dir_nr, const int files_per_dir,
	const char file_basename[],
	const int partition_numbers[], char dirnames[][MAXLEN],
	const int filesize, int create)
{
    for (int i = 0 ; i < dir_nr ; i++) {
	int part_num = partition_numbers[i] ;

	sprintf(dirnames[i], "%s%d", dir_basename, part_num) ;
	fprintf(stderr, "%s ", dirnames[i]) ;
	if (! create)
		continue ;

	for (int j = 0 ; j < files_per_dir ; j++) {
	    char filename[MAXLEN] ;

	    sprintf(filename, "%s/%s%d", 
		dirnames[i], file_basename, j) ;
	    check_rewrite_file(filename, filesize) ;
	}
    } // end of per-directory loop
    fprintf(stderr, "\n") ;
}

/*
 * Initialize the array of directory names and create/rewrite
 * non-existing/too short files
 */
void init_fileset(int create)
{
    int fd ;

    system("for ((i = 5 ; i < 35 ; i++)) ; do mount /mnt/hdb$i -o noatime; done") ;
   
    fprintf(stderr, "checking/creating html files:\n") ;
    check_create_files(HTML_DIR_NR, HTML_FILES_PER_DIR, html_file_basename,
	html_partition_numbers, html_dirnames, HTML_FILE_SIZE, create) ;

    fprintf(stderr, "checking/creating large files:\n") ;
    check_create_files(LARGE_DIR_NR, LARGE_FILES_PER_DIR, large_file_basename,
	large_partition_numbers, large_dirnames, LARGE_MAX_FILE_SIZE, create) ;

    fprintf(stderr, "creating fake logfile\n") ;
    fd = open(fakelog_filename, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR) ;
    if (fd == -1) {
	printf("could not open/create %s: %m\n", fakelog_filename);
	exit(1);
    }
    close(fd) ;

    fprintf(stderr, "unmount/mounting partitions WITH NOATIME\n") ;
    system("for ((i = 5 ; i < 35 ; i++)) ; do umount /mnt/hdb$i ; done") ;
    system("for ((i = 5 ; i < 35 ; i++)) ; do mount /mnt/hdb$i -o noatime ; done") ;
}

enum file_type_t {HTML, LARGE} ;

// return a random number between 0 and N-1
int gen_rand(const int N)
{
    return static_cast<int>( (rand() / (RAND_MAX + 1.0) ) * N) ;
}

void get_next_filename(file_type_t file_type, char filename[])
{
    const int dir_nr = file_type == HTML ? HTML_DIR_NR : LARGE_DIR_NR ;
    const int files_per_dir = file_type == HTML ? HTML_FILES_PER_DIR : 
	LARGE_FILES_PER_DIR ;

    const char *file_basename = file_type == HTML ? html_file_basename :
	large_file_basename ;
    const char (*dirnames)[MAXLEN] = file_type == HTML ? html_dirnames : 
	large_dirnames ;

    const int next_dir_idx = gen_rand(dir_nr) ;

    const int next_file_idx = gen_rand(files_per_dir) ;
    sprintf(filename, "%s/%s%d", 
		dirnames[next_dir_idx], file_basename, next_file_idx);
}

double time_since(struct timeval *start, struct timeval *end)
{
	double sec, usec, ret;

	sec = end->tv_sec - start->tv_sec;
	usec = end->tv_usec - start->tv_usec;
	if (sec > 0 && usec < 0) {
		sec = sec - 1;
		usec = usec + 1000000;
	}

	ret = sec + usec / (double)1000000;
	return ret > 0 ? ret : 0;
}

void read_file_print_stats(const file_type_t file_type, int idx, int num_iter,
	int file_idx)
{
	struct timeval start_tv, end_tv;
	unsigned char buffer[BLOCK_SIZE];
	long curr_iter, fsize = 0 ;
	double fsize_MiB ;
	double etime;
	char filename[MAXLEN] ;
	int fd ;
	const long max_iter = LARGE_MIN_FILE_SIZE/BLOCK_SIZE + 1 + 
	    gen_rand((LARGE_MAX_FILE_SIZE - LARGE_MIN_FILE_SIZE)/BLOCK_SIZE) ;

	get_next_filename(file_type, filename) ;
	fprintf(stderr, "opening file %s\n", filename) ;
	gettimeofday(&start_tv, NULL);
    	fd = open(filename, O_RDONLY);
    	if (fd == -1) {
	    printf("could not open %s: %m\n", filename);
	    exit(1);
  	}

	// read whole html file
	for (curr_iter = 0 ; file_type == HTML || curr_iter < max_iter ; 
	     curr_iter++) {
	    int ret = read(fd, buffer, BLOCK_SIZE) ;
	    if(max_iter < 10 || (curr_iter % (max_iter / 10)) == 0) 
		fprintf(stderr, "[process %d / iter %d / file %d (%s)] "
		"after read at iter %ld\n",
		idx, num_iter, file_idx, filename, curr_iter) ;
	    if (ret < 0) {
		printf("could not read from file: %m\n");
	 	exit(1) ;
	    }
	    if (ret > 0)
		fsize += ret ;
	    else
		break ;
	}
	fprintf(stderr, "[process %d / iter %d / file %d (%s)] "
		"before gettime\n",
		idx, num_iter, file_idx, filename) ;
	gettimeofday(&end_tv, NULL);
	etime = time_since(&start_tv, &end_tv);
	close(fd) ;
        fsize_MiB = fsize / 1024.0 / 1024.0 ; // convert to MiB
	fprintf(stderr, "[process %d / iter %d / file %d (%s)] read %lf MiB "
		"in %f sec (%fMiB/s)\n", 
		idx, num_iter, file_idx, filename,
		fsize_MiB, etime, fsize_MiB / etime);

	if (num_iter > 0)
	    if (file_type == HTML)
		fprintf(html_lat_log_fp, "%f\n", etime) ;
	    else
		fprintf(large_bw_log_fp, "%f\n", fsize_MiB / etime) ;
}

void worker(int idx)
{
    unsigned char append_buffer[MAX_APPEND_SIZE];
    int fd ;
    int append_size ; 

    srand(idx + 1 + time(NULL)) ;
    fprintf(stderr, "worker %d starting with %d\n", idx, rand()) ;

    for (int num_iter = 0 ; idx != PROCESS_NR - 1 || 
	    num_iter < MAX_PER_PROCESS_ITER ; num_iter++) {
	for (int i = 0 ; i < FILES_PER_PROCESS ; i++) {
	    if (gen_rand(PROB_SCALE) == PROB_SCALE - LARGE_FILE_PROB) 
		read_file_print_stats(LARGE, idx, num_iter, i) ;
	    else // read html file
		read_file_print_stats(HTML, idx, num_iter, i) ;
	} // end of outer loop

	fd = open(fakelog_filename, O_WRONLY | O_APPEND) ;
	if (fd == -1) {
	    printf("could not open %s: %m\n", fakelog_filename);
	    exit(1);
	}
	append_size = 1 + gen_rand(MAX_APPEND_SIZE) ;
	write(fd, append_buffer, append_size) ;
	close(fd) ;
    }

    return ;
}

void term_handler(int)
{
	fprintf(stderr, "term signal received\n") ;
	fflush(html_lat_log_fp) ;
	fflush(large_bw_log_fp) ;
	exit(0) ;
}

int main(int argc, char *argv[])
{
    int idx, create = 1 ;
    int fd, status ;
    char html_lat_log_filename[MAXLEN], large_bw_log_filename[MAXLEN] ;

    if (argc > 1)
	create = 0 ;
    
    system("cat /sys/block/hdb/queue/scheduler > /dev/stderr") ;
    init_fileset(create) ;
    
    fprintf(stderr, 
	"probability of large files: %f\n", 
	(double)LARGE_FILE_PROB/PROB_SCALE) ;
    fprintf(stderr, "creating processes\n") ;
    for (idx = 0; idx < PROCESS_NR - 1; ++idx) {
	int ret = fork() ;
	if (ret < 0) {
	    printf("could not create process %d: %m\n", idx);
	    exit(1);
	}
	if (ret == 0)
	    break ;
    }

    fprintf(stderr, "creating html latencies logfile\n") ;
    sprintf(html_lat_log_filename, "%s%d", html_lat_log_basefilename, idx) ;
    fd = open(html_lat_log_filename, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR) ;
    html_lat_log_fp = fdopen(fd, "w") ;
    if (html_lat_log_fp == NULL) {
	printf("could not open/create %s: %m\n", html_lat_log_filename);
	exit(1);
    }

    fprintf(stderr, "creating large bandwidths logfile\n") ;
    sprintf(large_bw_log_filename, "%s%d", large_bw_log_basefilename, idx) ;
    fd = open(large_bw_log_filename, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR) ;
    large_bw_log_fp = fdopen(fd, "w") ;
    if (large_bw_log_fp == NULL) {
	printf("could not open/create %s: %m\n", large_bw_log_filename);
	exit(1);
    }

    if (idx == PROCESS_NR - 1) {
    	system("iostat -d -m 1 > webserver_emu_agg_thr.rawlog &") ;
    	worker(idx) ;
        fprintf(stderr, "killing iostat ...\n") ;
    	system("killall iostat") ;
    	system("grep \"hdb \" webserver_emu_agg_thr.rawlog | "
		"awk \"{ print \\$3 }\" > webserver_emu_agg_thr.log") ;
    	system("rm webserver_emu_agg_thr.rawlog") ;
    	system("sed -i '1,30d' webserver_emu_agg_thr.log") ;
    	system("sed -i -n -e :a -e '1,30!{P;N;D;};N;ba' "
		"webserver_emu_agg_thr.log");
        fprintf(stderr, "killing all processes ...\n") ;
	kill(0, SIGCONT) ;
    	while(wait(&status) != -1)
		;
    } else {
    	signal(SIGCONT, term_handler); 
    	worker(idx) ;
    }
    fprintf(stderr, "process %d finished\n", idx) ;
    return 0;
}

